==> 1.5.3.tar.gz: lib/src/fragment/fragment_container.dart <==
import 'dart:collection';
import 'dart:convert';

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:flutterx_application/flutterx_application.dart';
import 'package:flutterx_live_data/flutterx_live_data.dart';
import 'package:flutterx_preferences/flutterx_preferences.dart';
import 'package:flutterx_utils/flutterx_utils.dart';

part 'fragment.dart';
part 'fragment_manager.dart';

class FragmentContainer extends StatefulWidget {
  final AxisDirection axisDirection;
  final bool primary;
  final Storage? storage;
  final FragmentEntry? initialFragment;
  final Widget Function(Fragment fragment) fragmentWrapper;
  final List<Fragment> fragments;

  @override
  FragmentContainerKey get key => super.key as FragmentContainerKey;

  const FragmentContainer({
    required FragmentContainerKey key,
    this.axisDirection = AxisDirection.right,
    this.primary = true,
    this.storage,
    this.initialFragment,
    this.fragmentWrapper = removePrimaryScrollController,
    required this.fragments,
  })  : assert(fragments.length > 0, 'fragments cannot be empty'),
        super(key: key);

  @override
  State<FragmentContainer> createState() => FragmentContainerState();

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) {
    super.debugFillProperties(properties);
    properties
      ..add(EnumProperty<AxisDirection>('axisDirection', axisDirection))
      ..add(DiagnosticsProperty<bool>('primary', primary))
      ..add(DiagnosticsProperty<Storage?>('storage', storage))
      ..add(ObjectFlagProperty<Widget Function(Fragment)>.has('fragmentWrapper', fragmentWrapper))
      ..add(DiagnosticsProperty<FragmentEntry?>('initialFragment', initialFragment));
  }

  static Widget removePrimaryScrollController(Fragment fragment) => PrimaryScrollController.none(child: fragment);

  static FragmentManager of(BuildContext context) => context.findAncestorStateOfType<FragmentContainerState>()!;

  static FragmentManager ofName(String name) => FragmentContainerKey(name).currentState!;
}

class FragmentContainerState extends State<FragmentContainer> implements FragmentManager {
  late PageController _controller;
  List<Fragment> _fragments = const [];
  List<Widget> _children = const [];
  Map<FragmentKey, int> _indexByKey = const {};
  ScrollDirection _direction = ScrollDirection.idle;

  @override
  FragmentContainer get container => widget;

  @override
  late FragmentStack currentFragment;

  @override
  LiveData<bool> get canGoBack => currentFragment.canUndo;

  @override
  LiveData<bool> get canGoForward => currentFragment.canRedo;

  @override
  void initState() {
    super.initState();
    _indexFragments();
    _initialize();
  }

  @override
  void didUpdateWidget(covariant FragmentContainer oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (_fragments != widget.fragments) _indexFragments();
    if (oldWidget.key != widget.key || !identical(oldWidget.storage, widget.storage)) _initialize(update: true);
  }

  @override
  Widget build(BuildContext context) {
    if (!currentFragment.initialized) {
      currentFragment.value = _initialKey(context);
      _initController();
    }
    return Scrollable(
        key: ValueKey(widget.key),
        axisDirection: widget.axisDirection,
        controller: _controller,
        physics: const NeverScrollableScrollPhysics(),
        restorationId: widget.key.name,
        scrollBehavior: ScrollConfiguration.of(context).copyWith(scrollbars: false),
        viewportBuilder: (context, position) =>
            Viewport(cacheExtent: 0, axisDirection: widget.axisDirection, offset: position, slivers: <Widget>[
              SliverFillViewport(
                  viewportFraction: _controller.viewportFraction, delegate: SliverChildListDelegate(_children)),
            ]));
  }

  @override
  void navigateTo<T extends Fragment>({ScrollDirection direction = ScrollDirection.idle, dynamic args}) {
    for (final item in _fragments) {
      if (item is T) {
        _direction = direction;
        currentFragment.value = FragmentEntry(item.key, args != null ? jsonEncode(args) : null);
        return;
      }
    }
    throw StateError('No fragment found with type $T');
  }

  @override
  void navigateToKey(FragmentKey key, {ScrollDirection direction = ScrollDirection.idle, dynamic args}) {
    _direction = direction;
    currentFragment.value = FragmentEntry(key, args != null ? jsonEncode(args) : null);
  }

  @override
  bool back({ScrollDirection direction = ScrollDirection.reverse}) {
    if (currentFragment.canUndo.value) {
      _direction = direction;
      currentFragment.historyUndo();
      return true;
    }
    return false;
  }

  @override
  bool forward({ScrollDirection direction = ScrollDirection.forward}) {
    if (currentFragment.canRedo.value) {
      _direction = direction;
      currentFragment.historyRedo();
      return true;
    }
    return false;
  }

  void _indexFragments() {
    _fragments = widget.fragments;
    _children = KeyedSubtree.ensureUniqueKeysForList(_fragments.map(widget.fragmentWrapper).toList(growable: false));
    var i = 0;
    _indexByKey = {for (final fragment in _fragments) fragment.key: i++};
  }

  void _initialize({bool update = false}) {
    if (update) currentFragment.removeObserver(_pushEntry);
    currentFragment = FragmentStack._(key: '${widget.key.name}_active_fragment', storage: widget.storage)
      ..observeForever(_pushEntry, dispatcher: Dispatcher.postFrame);
    if (currentFragment.initialized) _initController();
  }

  Future<void> _pushEntry(FragmentEntry entry) async {
    final currentKey = entry.key;
    final currentIndex = _indexByKey[currentKey];
    if (currentIndex == null) {
      currentFragment.value = _initialKey(context);
      return;
    }
    final lastIndex = _controller.page?.round();
    if (currentIndex == lastIndex) return;
    if (lastIndex == null) return _initController();
    final lastKey = _indexByKey.keys.elementAt(lastIndex);
    lastKey.currentState?.let((state) => state._active ? state.onPause() : null);
    currentKey.currentState?.setArguments(entry.args);
    final bool forward;
    switch (_direction) {
      case ScrollDirection.idle:
        forward = lastIndex < currentIndex;
        break;
      case ScrollDirection.forward:
        forward = true;
        break;
      case ScrollDirection.reverse:
        forward = false;
        break;
    }
    if ((forward && currentIndex - lastIndex == 1) || (!forward && currentIndex - lastIndex == -1)) {
      await _controller.animateToPage(currentIndex, duration: kThemeAnimationDuration, curve: Curves.ease);
    } else {
      final startIndex = forward
          ? currentIndex == 0
              ? 0
              : currentIndex - 1
          : currentIndex == _children.length - 1
              ? currentIndex
              : currentIndex + 1;
      final endIndex = forward
          ? currentIndex == 0
              ? 1
              : currentIndex
          : currentIndex == _children.length - 1
              ? currentIndex - 1
              : currentIndex;
      final originalChildren = _children;
      setState(() {
        _children = List<Widget>.of(_children, growable: false);
        final tmp = _children[startIndex];
        _children[startIndex] = _children[lastIndex];
        _children[lastIndex] = tmp;
      });
      _controller.jumpToPage(startIndex);
      await _controller.animateToPage(endIndex, duration: kThemeAnimationDuration, curve: Curves.ease);
      if (mounted) {
        setState(() => widget.fragments != _fragments ? _indexFragments() : _children = originalChildren);
        if (currentIndex != endIndex) _controller.jumpToPage(currentIndex);
      }
    }
    if (!mounted) return;
    if (widget.primary) AppPage.of(context).fragment = currentKey;
    currentKey.currentState?.let((state) => state._active ? null : state.onResume());
  }

  FragmentEntry _initialKey(BuildContext context) {
    if (widget.primary) {
      final key = AppPage.of(context).fragment;
      if (key != null && _indexByKey.containsKey(key)) return FragmentEntry(key);
    }
    final entry = widget.initialFragment;
    if (entry != null && _indexByKey.containsKey(entry)) return entry;
    return FragmentEntry(_fragments.first.key);
  }

  void _initController() =>
      _controller = PageController(initialPage: _indexByKey[currentFragment.value.key]!, keepPage: false);

  void _onFragmentAttached(FragmentState state) {
    final current = currentFragment.value;
    if (current.key == state.widget.key) {
      state.setArguments(current.args);
      if (!state._active) state.onResume();
    }
  }

  void _onFragmentDetached(FragmentState state) {
    final current = currentFragment.value;
    if (current.key == state.widget.key) if (state._active) state.onPause();
  }
}

@immutable
class FragmentEntry with DTO {
  final FragmentKey key;
  final String? args;

  const FragmentEntry(this.key, [this.args]);

  factory FragmentEntry.fromJson(JsonObject json) => FragmentEntry(FragmentKey(json['name']), json['args']);

  @override
  JsonObject toJson() => {'name': key.name, if (args != null) 'args': args};

  @override
  bool operator ==(Object other) => other is FragmentEntry && other.key == key && other.args == args;

  @override
  int get hashCode => hashValues(key, args);
}

class FragmentStack extends MutableLiveData<FragmentEntry> with HistoryMixin<FragmentEntry> {
  final String _indexKey;
  final Storage? _storage;
  final Queue<String>? _queue;

  @override
  Queue<String> get queue => _queue ?? super.queue;

  FragmentStack._({required String key, required Storage? storage})
      : _indexKey = '${key}_index',
        _storage = storage,
        _queue = storage == null ? null : StorageLiveQueue(key: key, storage: storage),
        super.late() {
    if (storage != null) {
      final index = storage[_indexKey]?.let(int.parse) ?? -1;
      if (index != -1) {
        historyIndex = index;
        restoreState();
      }
    }
  }

  @override
  void updateValues() {
    super.updateValues();
    final storage = _storage;
    if (storage != null) storage[_indexKey] = historyIndex.toString();
  }

  @override
  String serialize(FragmentEntry data) => jsonEncode(data.toJson());

  @override
  FragmentEntry deserialize(String data) => FragmentEntry.fromJson(jsonDecode(data));
}

class FragmentContainerKey extends GlobalKey<FragmentContainerState> {
  final String name;

  const FragmentContainerKey(this.name) : super.constructor();

  @override
  bool operator ==(Object other) => other is FragmentContainerKey && identical(other.name, name);

  @override
  int get hashCode => identityHashCode(name);

  @override
  String toString() => 'FragmentContainerKey($name)';
}

==> 1.5.7.tar.gz: lib/src/fragment/fragment_container.dart <==
import 'dart:collection';
import 'dart:convert';

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:flutterx_application/flutterx_application.dart';
import 'package:flutterx_live_data/flutterx_live_data.dart';
import 'package:flutterx_preferences/flutterx_preferences.dart';
import 'package:flutterx_utils/flutterx_utils.dart';

part 'fragment.dart';
part 'fragment_manager.dart';

class FragmentContainer extends StatefulWidget {
  final AxisDirection axisDirection;
  final bool primary;
  final Storage? storage;
  final FragmentEntry? initialFragment;
  final UnknownEntryStrategy unknownEntryStrategy;
  final Widget Function(Fragment fragment) fragmentWrapper;
  final List<Fragment> fragments;

  @override
  FragmentContainerKey get key => super.key as FragmentContainerKey;

  const FragmentContainer({
    required FragmentContainerKey key,
    this.axisDirection = AxisDirection.right,
    this.primary = true,
    this.storage,
    this.initialFragment,
    this.unknownEntryStrategy = UnknownEntryStrategy.idle,
    this.fragmentWrapper = removePrimaryScrollController,
    required this.fragments,
  })  : assert(fragments.length > 0, 'fragments cannot be empty'),
        super(key: key);

  @override
  State<FragmentContainer> createState() => FragmentContainerState();

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) {
    super.debugFillProperties(properties);
    properties
      ..add(EnumProperty<AxisDirection>('axisDirection', axisDirection))
      ..add(DiagnosticsProperty<bool>('primary', primary))
      ..add(DiagnosticsProperty<Storage?>('storage', storage))
      ..add(DiagnosticsProperty<FragmentEntry?>('initialFragment', initialFragment))
      ..add(EnumProperty<UnknownEntryStrategy>('unknownEntryStrategy', unknownEntryStrategy))
      ..add(ObjectFlagProperty<Widget Function(Fragment)>.has('fragmentWrapper', fragmentWrapper));
  }

  static Widget removePrimaryScrollController(Fragment fragment) => PrimaryScrollController.none(child: fragment);

  static FragmentManager of(BuildContext context) => context.findAncestorStateOfType<FragmentContainerState>()!;

  static FragmentManager ofName(String name) => FragmentContainerKey(name).currentState!;
}

class FragmentContainerState extends State<FragmentContainer> implements FragmentManager {
  late PageController _controller;
  List<Fragment> _fragments = const [];
  List<Widget> _children = const [];
  Map<FragmentKey, int> _indexByKey = const {};
  ScrollDirection _direction = ScrollDirection.idle;

  @override
  FragmentContainer get container => widget;

  @override
  final FragmentStack currentFragment = FragmentStack._();

  @override
  LiveData<bool> get canGoBack => currentFragment.canUndo;

  @override
  LiveData<bool> get canGoForward => currentFragment.canRedo;

  @override
  void initState() {
    super.initState();
    _indexFragments();
    currentFragment
      .._initialize(key: widget.key, storage: widget.storage)
      ..observeForever(_pushEntry, dispatcher: Dispatcher.postFrame);
  }

  @override
  void didUpdateWidget(covariant FragmentContainer oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (!listEquals(_fragments, widget.fragments)) _indexFragments();
    if (oldWidget.key != widget.key || !identical(oldWidget.storage, widget.storage))
      currentFragment._initialize(key: widget.key, storage: widget.storage);
  }

  @override
  Widget build(BuildContext context) {
    if (!_ready && !_initController()) return const SizedBox();
    return Scrollable(
        key: ValueKey(widget.key),
        axisDirection: widget.axisDirection,
        controller: _controller,
        physics: const NeverScrollableScrollPhysics(),
        restorationId: widget.key.name,
        scrollBehavior: ScrollConfiguration.of(context).copyWith(scrollbars: false),
        viewportBuilder: (context, position) =>
            Viewport(cacheExtent: 0, axisDirection: widget.axisDirection, offset: position, slivers: <Widget>[
              SliverFillViewport(
                  viewportFraction: _controller.viewportFraction, delegate: SliverChildListDelegate(_children)),
            ]));
  }

  @override
  void dispose() {
    currentFragment.removeObserver(_pushEntry);
    super.dispose();
  }

  @override
  void navigateTo<T extends Fragment>({ScrollDirection direction = ScrollDirection.idle, dynamic args}) {
    for (final item in _fragments) {
      if (item is T) {
        _direction = direction;
        currentFragment.value = FragmentEntry(item.key, args != null ? jsonEncode(args) : null);
        return;
      }
    }
    throw StateError('No fragment found with type $T');
  }

  @override
  void navigateToKey(FragmentKey key, {ScrollDirection direction = ScrollDirection.idle, dynamic args}) {
    _direction = direction;
    currentFragment.value = FragmentEntry(key, args != null ? jsonEncode(args) : null);
  }

  @override
  bool back({ScrollDirection direction = ScrollDirection.reverse}) {
    if (currentFragment.canUndo.value) {
      _direction = direction;
      currentFragment.historyUndo();
      return true;
    }
    return false;
  }

  @override
  bool forward({ScrollDirection direction = ScrollDirection.forward}) {
    if (currentFragment.canRedo.value) {
      _direction = direction;
      currentFragment.historyRedo();
      return true;
    }
    return false;
  }

  void _indexFragments() {
    _fragments = widget.fragments.toList(growable: false);
    _children = KeyedSubtree.ensureUniqueKeysForList(_fragments.map(widget.fragmentWrapper).toList(growable: false));
    var i = 0;
    _indexByKey = {for (final fragment in _fragments) fragment.key: i++};
  }

  Future<void> _pushEntry(FragmentEntry entry) async {
    final currentKey = entry.key;
    final currentIndex = _indexByKey[currentKey];
    if (currentIndex == null) {
      switch (widget.unknownEntryStrategy) {
        case UnknownEntryStrategy.snap:
          currentFragment.value = _initialEntry(context);
          return;
        case UnknownEntryStrategy.crash:
          throw StateError('Invalid entry: $entry, available entries: [${_fragments.join(',')}]');
        case UnknownEntryStrategy.idle:
          return;
      }
    }
    final lastIndex = _ready ? _controller.page?.round() : null;
    if (currentIndex == lastIndex) return;
    if (lastIndex == null) {
      _initController(currentIndex);
      return;
    }
    final lastKey = _indexByKey.keys.elementAt(lastIndex);
    lastKey.currentState?.let((state) => state._active ? state.onPause() : null);
    currentKey.currentState?.setArguments(entry.args);
    final bool forward;
    switch (_direction) {
      case ScrollDirection.idle:
        forward = lastIndex < currentIndex;
        break;
      case ScrollDirection.forward:
        forward = true;
        break;
      case ScrollDirection.reverse:
        forward = false;
        break;
    }
    if ((forward && currentIndex - lastIndex == 1) || (!forward && currentIndex - lastIndex == -1)) {
      await _controller.animateToPage(currentIndex, duration: kThemeAnimationDuration, curve: Curves.ease);
    } else {
      final startIndex = forward
          ? currentIndex == 0
              ? 0
              : currentIndex - 1
          : currentIndex == _children.length - 1
              ? currentIndex
              : currentIndex + 1;
      final endIndex = forward
          ? currentIndex == 0
              ? 1
              : currentIndex
          : currentIndex == _children.length - 1
              ? currentIndex - 1
              : currentIndex;
      final originalChildren = _children;
      setState(() {
        _children = List<Widget>.of(_children, growable: false);
        final tmp = _children[startIndex];
        _children[startIndex] = _children[lastIndex];
        _children[lastIndex] = tmp;
      });
      _controller.jumpToPage(startIndex);
      await _controller.animateToPage(endIndex, duration: kThemeAnimationDuration, curve: Curves.ease);
      if (mounted) {
        setState(() => listEquals(widget.fragments, _fragments) ? _children = originalChildren : _indexFragments());
        if (currentIndex != endIndex) _controller.jumpToPage(currentIndex);
      }
    }
    if (!mounted) return;
    if (widget.primary) AppPage.of(context).fragment = currentKey;
    currentKey.currentState?.let((state) => state._active ? null : state.onResume());
  }

  FragmentEntry _initialEntry(BuildContext context) {
    if (widget.primary) {
      final key = AppPage.of(context).fragment;
      if (key != null && _indexByKey.containsKey(key)) return FragmentEntry(key);
    }
    final entry = widget.initialFragment;
    if (entry != null && _indexByKey.containsKey(entry)) return entry;
    return FragmentEntry(_indexByKey.keys.first);
  }

  bool _ready = false;

  bool _initController([int? initialIndex]) {
    if (initialIndex == null) {
      final entry =
          currentFragment.initialized ? currentFragment.value : currentFragment.value = _initialEntry(context);
      initialIndex = _indexByKey[entry.key];
      if (initialIndex == null) {
        switch (widget.unknownEntryStrategy) {
          case UnknownEntryStrategy.snap:
            currentFragment.value = _initialEntry(context);
            initialIndex = _indexByKey[entry.key];
            if (initialIndex == null)
              throw StateError(
                  'Failed to determinate initial index for entry: $entry, available entries: [${_fragments.join(',')}]');
            break;
          case UnknownEntryStrategy.crash:
            throw StateError('Invalid entry: $entry, available entries: [${_fragments.join(',')}]');
          case UnknownEntryStrategy.idle:
            return false;
        }
      }
    }
    _controller = PageController(initialPage: initialIndex, keepPage: false);
    return _ready = true;
  }

  void _onFragmentAttached(FragmentState state) {
    final current = currentFragment.value;
    if (current.key == state.widget.key) {
      state.setArguments(current.args);
      if (!state._active) state.onResume();
    }
  }

  void _onFragmentDetached(FragmentState state) {
    final current = currentFragment.value;
    if (current.key == state.widget.key) if (state._active) state.onPause();
  }
}

@immutable
class FragmentEntry with DTO {
  final FragmentKey key;
  final String? args;

  const FragmentEntry(this.key, [this.args]);

  factory FragmentEntry.fromJson(JsonObject json) => FragmentEntry(FragmentKey(json['name']), json['args']);

  @override
  JsonObject toJson() => {'name': key.name, if (args != null) 'args': args};

  @override
  bool operator ==(Object other) => other is FragmentEntry && other.key == key && other.args == args;

  @override
  int get hashCode => hashValues(key, args);
}

class FragmentStack extends MutableLiveData<FragmentEntry> with HistoryMixin<FragmentEntry> {
  late Queue<String> _queue;
  late String _indexKey;
  Storage? _storage;

  @override
  Queue<String> get queue => _queue;

  FragmentStack._() : super.late();

  void _initialize({required FragmentContainerKey key, required Storage? storage}) {
    _queue = storage == null ? Queue() : StorageLiveQueue(key: '${key.name}_active_fragment', storage: storage);
    _indexKey = '${key.name}_active_fragment_index';
    _storage = storage;
    if (storage != null) {
      final index = storage[_indexKey]?.let(int.parse) ?? -1;
      if (index != -1) {
        historyIndex = index;
        restoreState();
      }
    }
  }

  @override
  void updateValues() {
    super.updateValues();
    final storage = _storage;
    if (storage != null) storage[_indexKey] = historyIndex.toString();
  }

  @override
  String serialize(FragmentEntry data) => jsonEncode(data.toJson());

  @override
  FragmentEntry deserialize(String data) => FragmentEntry.fromJson(jsonDecode(data));
}

class FragmentContainerKey extends GlobalKey<FragmentContainerState> {
  final String name;

  const FragmentContainerKey(this.name) : super.constructor();

  @override
  bool operator ==(Object other) => other is FragmentContainerKey && identical(other.name, name);

  @override
  int get hashCode => identityHashCode(name);

  @override
  String toString() => 'FragmentContainerKey($name)';
}

enum UnknownEntryStrategy {
  /// Fallback to initial entry
  snap,

  /// Throw an error
  crash,

  /// Do nothing until the entry is available or another entry is provided
  idle,
}
